import typing

from .scan import scan
from .resolve import resolve
from .util import tree, order
from .types import CallableInfo
from .inject import inject, ainject


def from_(call: typing.Callable[..., typing.Any]) -> CallableInfo:
    """
    Use callable as dependency for parameter of function(alias for fundi.scan.scan)
    :param call: callable to be used as dependency
    :return: callable information
    """
    return scan(call)



import typing
from dataclasses import dataclass

__all__ = ["R", "Parameter", "CallableInfo", "ParameterResult"]

R = typing.TypeVar("R")


@dataclass
class Parameter:
    name: str
    annotation: type
    from_: "CallableInfo | None"


@dataclass
class CallableInfo(typing.Generic[R]):
    call: typing.Callable[..., typing.Any]
    async_: bool
    generator: bool
    parameters: list[Parameter]


@dataclass
class ParameterResult:
    parameter_name: str
    value: typing.Any | None
    dependency: CallableInfo | None
    resolved: bool



import typing

from fundi.types import CallableInfo, ParameterResult


def resolve(
    scope: typing.Mapping[str, typing.Any],
    info: CallableInfo,
    cache: typing.Mapping[typing.Callable, typing.Any],
) -> typing.Generator[ParameterResult, None, None]:
    """
    Try to resolve values from cache or scope for callable parameters

    Recommended use case:

    Example::

        values = {}
        cache = {}
        for result in resolve(scope, info, cache):
            value = result.value
            name = result.parameter_name

            if not result.resolved:
                value = inject(scope, info, stack, cache)
                cache[name] = value

            values[name] = value


    :param scope: container with contextual values
    :param info: callable information
    :param cache: solvation cache(modify it if necessary while resolving)
    :return: generator with solvation results
    """
    for parameter in info.parameters:
        if parameter.from_:
            dependency = parameter.from_
            call = dependency.call

            if call in cache:
                yield ParameterResult(parameter.name, cache[call], dependency, resolved=True)
            else:
                yield ParameterResult(parameter.name, None, dependency, resolved=False)
        else:
            if parameter.name not in scope:
                raise ValueError(
                    f"Cannot resolve {parameter.name} for {info.call} - Scope does not contain required value"
                )
            yield ParameterResult(parameter.name, scope[parameter.name], None, resolved=True)



import typing
from contextlib import ExitStack, AsyncExitStack

from fundi.resolve import resolve
from fundi.types import CallableInfo
from fundi.util import _call_sync, _call_async


def inject(
    scope: typing.Mapping[str, typing.Any],
    info: CallableInfo[typing.Any],
    stack: ExitStack,
    cache: typing.Mapping[typing.Callable, typing.Any] = None,
) -> typing.Any:
    """
    Synchronously inject dependencies into callable.

    :param scope: container with contextual values
    :param info: callable information
    :param stack: exit stack to properly handle generator dependencies
    :param cache: dependency cache
    :return: result of callable
    """
    if info.async_:
        raise RuntimeError("Cannot process async functions in synchronous injection")

    if cache is None:
        cache = {}

    values = {}

    for result in resolve(scope, info, cache):
        name = result.parameter_name
        value = result.value

        if not result.resolved:
            dependency = result.dependency

            value = inject(scope, dependency, stack, cache)
            cache[dependency.call] = value

        values[name] = value

    return _call_sync(stack, info, values)


async def ainject(
    scope: typing.Mapping[str, typing.Any],
    info: CallableInfo[typing.Any],
    stack: AsyncExitStack,
    cache: typing.Mapping[typing.Callable, typing.Any] = None,
) -> typing.Any:
    """
    Asynchronously inject dependencies into callable.

    :param scope: container with contextual values
    :param info: callable information
    :param stack: exit stack to properly handle generator dependencies
    :param cache: dependency cache
    :return: result of callable
    """
    if cache is None:
        cache = {}

    values = {}

    for result in resolve(scope, info, cache):
        name = result.parameter_name
        value = result.value

        if not result.resolved:
            dependency = result.dependency

            value = await ainject(scope, dependency, stack, cache)
            cache[dependency.call] = value

        values[name] = value

    if not info.async_:
        return _call_sync(stack, info, values)

    return await _call_async(stack, info, values)



import typing
from typing import overload
from contextlib import ExitStack, AsyncExitStack

from fundi.types import CallableInfo

R = typing.TypeVar("R")

ExitStack = AsyncExitStack | ExitStack

@overload
def inject(
    scope: typing.Mapping[str, typing.Any],
    info: CallableInfo[typing.Generator[R, None, None]],
    stack: ExitStack,
    cache: typing.Mapping[typing.Callable, typing.Any] = None,
) -> R: ...
@overload
def inject(
    scope: typing.Mapping[str, typing.Any],
    info: CallableInfo[R],
    stack: ExitStack,
    cache: typing.Mapping[typing.Callable, typing.Any] = None,
) -> R: ...
@overload
async def ainject(
    scope: typing.Mapping[str, typing.Any],
    info: CallableInfo[typing.Generator[R, None, None]],
    stack: AsyncExitStack,
    cache: typing.Mapping[typing.Callable, typing.Any] = None,
) -> R: ...
@overload
async def ainject(
    scope: typing.Mapping[str, typing.Any],
    info: CallableInfo[typing.AsyncGenerator[R, None]],
    stack: AsyncExitStack,
    cache: typing.Mapping[typing.Callable, typing.Any] = None,
) -> R: ...
@overload
async def ainject(
    scope: typing.Mapping[str, typing.Any],
    info: CallableInfo[typing.Awaitable[R]],
    stack: AsyncExitStack,
    cache: typing.Mapping[typing.Callable, typing.Any] = None,
) -> R: ...
@overload
async def ainject(
    scope: typing.Mapping[str, typing.Any],
    info: CallableInfo[R],
    stack: AsyncExitStack,
    cache: typing.Mapping[typing.Callable, typing.Any] = None,
) -> R: ...



import typing
from contextlib import AsyncExitStack, ExitStack

from fundi.resolve import resolve
from fundi.types import CallableInfo


def _call_sync(
    stack: ExitStack | AsyncExitStack,
    info: CallableInfo[typing.Any],
    values: typing.Mapping[str, typing.Any],
) -> typing.Any:
    """
    Synchronously call dependency callable.

    :param stack: exit stack to properly handle generator dependencies
    :param info: callable information
    :param values: callable arguments
    :return: callable result
    """
    value = info.call(**values)

    if info.generator:
        generator = value
        value = next(generator)

        def exit_generator():
            try:
                next(generator)
            except StopIteration:
                pass

        stack.callback(exit_generator)

    return value


async def _call_async(
    stack: AsyncExitStack, info: CallableInfo[typing.Any], values: typing.Mapping[str, typing.Any]
) -> typing.Any:
    """
    Asynchronously call dependency callable.

    :param stack: exit stack to properly handle generator dependencies
    :param info: callable information
    :param values: callable arguments
    :return: callable result
    """
    value = info.call(**values)

    if info.generator:
        generator = value
        value = await anext(generator)

        async def exit_generator():
            try:
                await anext(generator)
            except StopAsyncIteration:
                pass

        stack.push_async_callback(exit_generator)

    else:
        value = await value

    return value


def tree(
    scope: typing.Mapping[str, typing.Any],
    info: CallableInfo,
    cache: typing.Mapping[typing.Callable, typing.Mapping[str, typing.Any]] = None,
) -> typing.Mapping[str, typing.Any]:
    """
    Get tree of dependencies of callable.

    :param scope: container with contextual values
    :param info: callable information
    :param cache: tree generation cache
    :return: Tree of dependencies
    """
    if cache is None:
        cache = {}

    values = {}

    for result in resolve(scope, info, cache):
        name = result.parameter_name
        value = result.value

        if not result.resolved:
            value = tree(scope, result.dependency, cache)
            cache[result.dependency.call] = value

        values[name] = value

    return {"call": info.call, "values": values}


def order(
    scope: typing.Mapping[str, typing.Any],
    info: CallableInfo[typing.Any],
    cache: typing.Mapping[typing.Callable, list[typing.Callable]] = None,
) -> list[typing.Callable]:
    """
    Get resolving order of callable dependencies.

    :param info: callable information
    :param scope: container with contextual values
    :param cache: solvation cache
    :return: order of dependencies
    """
    if cache is None:
        cache = {}

    order_ = []

    for result in resolve(scope, info, cache):
        if not result.resolved:
            value = order(scope, result.dependency, cache)
            order_.extend(value)
            order_.append(result.dependency.call)
            cache[result.dependency.call] = value

    return order_



import typing
import inspect

from fundi.types import R, CallableInfo, Parameter


def scan(call: typing.Callable[..., R]) -> CallableInfo[R]:
    """
    Get callable information
    :param call: callable to get information from
    :return: callable information
    """
    params = []

    for param in inspect.signature(call).parameters.values():
        if isinstance(param.default, CallableInfo):
            params.append(Parameter(param.name, param.annotation, param.default))
            continue

        params.append(Parameter(param.name, param.annotation, None))

    async_ = inspect.iscoroutinefunction(call) or inspect.isasyncgenfunction(call)
    generator = inspect.isgeneratorfunction(call) or inspect.isasyncgenfunction(call)

    return CallableInfo(call=call, async_=async_, generator=generator, parameters=params)
